#include "stdafx.h"
#include "init.h"
#include "ioctls.h"
#include "dispatch.h"
#include "list.h"

// note that the explicit status codes used in the initialization
// parts of the driver aren't my choice, dheapmon.exe relies 
// on them being what they are to print out an appropiate error message.
// It doesn't use formatmessage to get the strings
// so changing them for ones with a more representative
// error text isn't an option

extern "C"
{
	DRIVER_INITIALIZE DriverEntry;
	DRIVER_UNLOAD DHeapDmpUnload;
}

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGE, DHeapDmpUnload)
#endif

NTSTATUS DriverEntry(PDRIVER_OBJECT pDrvObj, PUNICODE_STRING pusRegPath)
{
	MmPageEntireDriver(&DriverEntry);
	DBG_TEXT("In DriverEntry");

	UNICODE_STRING usDevName = RTL_CONSTANT_STRING(L"\\Device\\DHEAPDMP");

	// first, check for compatible OS version
	RTL_OSVERSIONINFOW osVersion = {sizeof(osVersion)};
	// never fails
	RtlGetVersion(&osVersion);
	// saved so we can correctly resolve the position of the desktop heap
	// in winstadesk.cpp
	g_dwMajorVersion = osVersion.dwMajorVersion;
	g_dwMinorVersion = osVersion.dwMinorVersion;
	
	// if the major version isn't 6, or the minor version is greater than 1 then fail
	if((g_dwMajorVersion != 6) || (g_dwMinorVersion > 1))
	{
		DBG_TEXT(
			"Expecting Windows version %d with a minor of %d at most "
			"but found version %d.%d",
			g_dwMajorVersionRequired,
			g_dwHighestMinorVersion,
			g_dwMajorVersion,
			g_dwMinorVersion
			);
		return STATUS_NOT_SUPPORTED;
	}

	// initialize the environment by reading our parameters from the registry
	NTSTATUS stat = InitializeEnvironment(pusRegPath);
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("Failed to initialize environment (0x%x)", stat);
		return stat;
	}

	// init the relevant function pointers
	pDrvObj->DriverUnload = &DHeapDmpUnload;
	pDrvObj->MajorFunction[IRP_MJ_CREATE] = &DHeapDmpDispatchCreate;
	pDrvObj->MajorFunction[IRP_MJ_CLOSE] = &DHeapDmpDispatchClose;
	pDrvObj->MajorFunction[IRP_MJ_DEVICE_CONTROL] = &DHeapDmpDispatchControl;

	// create our device object 
	PDEVICE_OBJECT pDevObj = NULL;
	stat = IoCreateDevice(pDrvObj, 0, &usDevName, FILE_DEVICE_UNKNOWN, 0, FALSE, &pDevObj);
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("Failed to create device. Error = 0x%x, Name = %wZ\n", stat, &usDevName);
		return stat;
	}

	// and the UM visible symbolic link
	stat = IoCreateSymbolicLink(&g_usSymName, &usDevName);
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("Failed to create symbolic link. Error = 0x%x, Name = %wZ\n", stat, &g_usSymName);
		IoDeleteDevice(pDevObj);
		return stat;
	}
	// and finally init our mutex
	KeInitializeGuardedMutex(&g_Mutex);
	return STATUS_SUCCESS;
}

void DHeapDmpUnload(PDRIVER_OBJECT pDrvObj)
{
	// just undo the damage we caused
	PAGED_CODE();
	DBG_TEXT("Unloading");
	IoDeleteSymbolicLink(&g_usSymName);
	FastMutexLock lock(&g_Mutex);
	EmptyDeskList();
	IoDeleteDevice(pDrvObj->DeviceObject);
}
